from dexy.commands.utils import init_wrapper
from dexy.utils import file_exists
import dexy.load_plugins
import dexy.reporter
import http.server
import os
import socket
import socketserver
import sys

NO_OUTPUT_MSG = """Please run dexy first, or specify a directory to serve. \
For help run 'dexy help -on serve'"""

class SimpleHTTPAuthRequestHandler(http.server.SimpleHTTPRequestHandler):
    def send_head(self):
        """Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is either a file object (which has to be copied
        to the outputfile by the caller unless the command was HEAD,
        and must be closed by the caller under all circumstances), or
        None, in which case the caller has nothing further to do.

        """
        if self.headers.getheader('Authorization') == None:
            self.send_response(401)
            self.send_header('WWW-Authenticate', 'Basic realm="%s"' % self.__class__.realm)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write("no authorization received")

        elif self.headers.getheader('Authorization') != "Basic %s" % self.__class__.authcode:
            self.send_response(401)
            self.send_header('WWW-Authenticate', 'Basic realm="%s"' % self.__class__.realm)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write("not authenticated")

        else:
            path = self.translate_path(self.path)
            f = None
            if os.path.isdir(path):
                if not self.path.endswith('/'):
                    # redirect browser - doing basically what apache does
                    self.send_response(301)
                    self.send_header("Location", self.path + "/")
                    self.end_headers()
                    return None
                for index in "index.html", "index.htm":
                    index = os.path.join(path, index)
                    if os.path.exists(index):
                        path = index
                        break
                else:
                    return self.list_directory(path)
            ctype = self.guess_type(path)
            try:
                # Always read in binary mode. Opening files in text mode may cause
                # newline translations, making the actual size of the content
                # transmitted *less* than the content-length!
                f = open(path, 'rb')
            except IOError:
                self.send_error(404, "File not found")
                return None
            self.send_response(200)
            self.send_header("Content-type", ctype)
            fs = os.fstat(f.fileno())
            self.send_header("Content-Length", str(fs[6]))
            self.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
            self.end_headers()
            return f

def serve_command(
        port=-1,
        reporters=['ws', 'output'], # Reporters whose output to try to serve (in order).
        username='', # http auth username to use (if provided)
        password='', # http auth password to use (if provided)
        realm='Dexy', # http auth realm to use (if username and password are provided)
        directory=False, # Custom directory to be served.
        **kwargs
        ):
    """
    Runs a simple web server on dexy-generated files.
    
    Will look first to see if the Website Reporter has run, if so this content
    is served. If not the standard output/ directory contents are served. You
    can also specify another directory to be served. The port defaults to 8085,
    this can also be customized. If a username and password are provided, uses
    HTTP auth to access pages.
    """

    if not directory:
        wrapper = init_wrapper(locals(), True)

        for alias in reporters:
            report_dir = dexy.reporter.Reporter.create_instance(alias).setting("dir")
            print("report dir", report_dir)
            if report_dir and file_exists(report_dir):
                directory = report_dir
                break

    if not directory:
        print(NO_OUTPUT_MSG)
        sys.exit(1)

    os.chdir(directory)

    if port < 0:
        ports = range(8085, 8100)
    else:
        ports = [port]

    p = None
    for p in ports:
        try:
            if username and password:
                import base64
                authcode = base64.b64encode("%s:%s" % (username, password))
                Handler = SimpleHTTPAuthRequestHandler
                Handler.authcode = authcode
                Handler.realm = realm
            else:
                Handler = http.server.SimpleHTTPRequestHandler
            httpd = socketserver.TCPServer(("", p), Handler)
        except socket.error:
            print("port %s already in use" % p)
            p = None
        else:
            break

    if p:
        print("serving contents of %s on http://localhost:%s" % (directory, p))
        if username and password and Handler.authcode:
            print("username '%s' and password '%s' are required to access contents" % (username, password))
        print("type ctrl+c to stop")
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            sys.exit(1)
    else:
        print("could not find a free port to serve on, tried", ports)
        sys.exit(1)



from dexy.doc import Doc
from tests.utils import assert_in_output
from tests.utils import assert_output
from tests.utils import wrap
from nose.exc import SkipTest

JAVA_SRC = """public class hello {
  public static void main(String args[]) {
    System.out.println("Java Hello World!");
  }
}"""

def test_javac_filter():
    # not using runfilter() because file has to be named 'hello.java'
    with wrap() as wrapper:
        doc = Doc("hello.java|javac",
                wrapper,
                [],
                contents=JAVA_SRC)
        wrapper.run_docs(doc)
        assert doc.output_data().is_cached()

def test_java_filter():
    # not using runfilter() because file has to be named 'hello.java'
    with wrap() as wrapper:
        doc = Doc("hello.java|java",
                wrapper,
                [],
                contents=JAVA_SRC)
        wrapper.run_docs(doc)
        assert str(doc.output_data()) == "Java Hello World!\n"

def test_jruby_filter():
    assert_output('jruby', "puts 1+1", "2\n")

def test_jirb_filter():
    assert_in_output('jirb', "puts 1+1",  ">> puts 1+1")

def test_jython_filter():
    assert_output('jython', "print 1+1", "2\n")

def test_jythoni_filter():
    raise SkipTest()
    assert_in_output('jythoni', "print 1+1",  ">>> print 1+1")



from dexy.exceptions import UserFeedback, InternalDexyProblem
from dexy.filters.pyg import PygmentsFilter
from pygments import highlight
import ply.lex as lex
import ply.yacc as yacc

class LexError(InternalDexyProblem):
    pass

class ParseError(UserFeedback):
    pass

tokens = (
    'AMP',
    'AT',
    'CODE',
    'COLONS',
    'DBLQUOTE',
    'END',
    'IDIOCLOSE',
    'IDIOOPEN',
    'EXP',
    'IDIO',
    'NEWLINE',
    'SGLQUOTE',
    'WHITESPACE',
    'WORD',
)

states = (
    ('idiostart', 'exclusive',),
    ('idio', 'exclusive',),
)

class Id(PygmentsFilter):
    """
    Splits files into sections based on comments like ### "foo"

    Replacement for idiopidae. Should be fully backwards-compatible.

    For more information about the settings starting with ply-, see the PLY
    YACC parser documentation http://www.dabeaz.com/ply/ply.html#ply_nn36
    """
    aliases = ['idio', 'id', 'idiopidae', 'htmlsections']
    _settings = {
            'examples' : ['idio', 'htmlsections'],
            'highlight' : ("Whether to apply syntax highlighting to sectional output.", None),
            'skip-extensions' : ("Because |idio gets applied to *.*, need to make it easy to skip non-textual files.", (".odt")),
            'remove-leading' : ("If a document starts with empty section named '1', remove it.", False),
            'ply-optimize' : ("Whether to use optimized mode for the lexer.", 1),
            'ply-write-tables' : ("Whether to generate parser table files (which will be stored in ply-outputdir and named ply-tabmodule).", 1),
            'ply-outputdir' : ("Location relative to where you run dexy in which ply will store table files. Defaults to dexy's log directory.", None),
            'ply-parsetab' : ("Name of parser tabfile (.py extension will be added) to be stored in ply-outputdir if ply-write-tables set to 1.", 'id_parsetab'),
            'ply-lextab' : ("Name of lexer tabfile (.py extension will be added) to be stored in ply-outputdir if ply-optimize is set to 1.", 'id_lextab'),
            'output-extensions' : PygmentsFilter.MARKUP_OUTPUT_EXTENSIONS + PygmentsFilter.IMAGE_OUTPUT_EXTENSIONS
            }

    def process(self):
        try:
            input_text = str(self.input_data)
        except UnicodeDecodeError:
            self.output_data['1'] = "non textual"
            self.output_data.save()
            return

        lexer.outputdir = self.setting('ply-outputdir')
        lexer.errorlog = self.doc.wrapper.log
        lexer.remove_leading = self.setting('remove-leading')
        parser.outputdir = self.setting('ply-outputdir')
        parser.errorlog = self.doc.wrapper.log
        parser.write_tables = self.setting('ply-write-tables')

        _lexer = lexer.clone()
        _lexer.sections = []
        _lexer.level = 0
        start_new_section(_lexer, 0, 0, _lexer.level)

        parser.parse(input_text + "\n", lexer=_lexer)
        strip_trailing_newline(_lexer)
        parser_output = _lexer.sections

        pyg_lexer = self.create_lexer_instance()
        pyg_formatter = self.create_formatter_instance()

        # TODO fix file extension if highlight is set to false
        do_highlight = self.setting('highlight')
        if do_highlight is None:
            if self.alias in ('htmlsections',):
                do_highlight = False
                if self.output_data.setting('canonical-output') is None:
                    self.output_data.update_settings({'canonical-output' : True})
            else:
                do_highlight = True

        for section in parser_output:
            if do_highlight:
                section['contents'] = highlight(section['contents'], pyg_lexer, pyg_formatter)
            self.output_data._data.append(section)
        self.output_data.save()

def t_error(t):
    raise LexError("Problem lexing at position %s." % t.lexpos)

def t_idio_error(t):
    print("comment '%s'" % t.lexer.lexdata[t.lexer.comment_start_pos:])
    print("all '%s'" % t.lexer.lexdata)
    print("char '%s'" % t.lexer.lexdata[t.lexpos-1:t.lexer.lexpos])
    raise LexError("Problem lexing in 'idio' state at position %s." % t.lexpos)

def t_idiostart_error(t):
    raise LexError("Problem lexing in 'idiostart' state at position %s." % t.lexpos)
        
def append_text(lexer, code):
    """
    Append to the currently active section.
    """
    set_current_section_contents(lexer, current_section_contents(lexer) + code)

def current_section_exists(lexer):
    return len(lexer.sections) > 0

def current_section_empty(lexer):
    return len(current_section_contents(lexer)) == 0

def current_section_contents(lexer):
    return lexer.sections[-1]['contents']

def set_current_section_contents(lexer, text):
    lexer.sections[-1]['contents'] = text

def strip_trailing_newline(lexer):
    set_current_section_contents(lexer, current_section_contents(lexer).rsplit("\n",1)[0])

def start_new_section(lexer, position, lineno, new_level, name=None):
    if name:
        if lexer.remove_leading:
            if len(lexer.sections) == 1 and current_section_empty(lexer):
                lexer.sections = []
    else:
        # Generate anonymous section name.
        name = str(len(lexer.sections)+1)

    try:
        change_level(lexer, new_level)
    except Exception:
        print(name)
        raise

    lexer.sections.append({
            'name' : name.rstrip(),
            'position' : position,
            'lineno' : lineno,
            'contents' : '',
            'level' : lexer.level
            })

def change_level(lexer, new_level):
    if new_level == lexer.level:
        pass
    elif new_level < lexer.level:
        pass
    elif new_level == lexer.level + 1:
        pass
    elif new_level > (lexer.level + 1):
        msg = "attempting to indent more than 1 level to %s from previous level %s"
        msgargs = (new_level, lexer.level)
        raise Exception(msg % msgargs)
    elif new_level < 0:
        raise Exception("attepmting to indent to level below 0, does not exist")
    else:
        msg = "logic error! new level %s current level %s"
        msgargs = (new_level, lexer.level)
        raise Exception(msg % msgargs)

    lexer.level = new_level

def next_char(t):
    return t.lexer.lexdata[t.lexer.lexpos:t.lexer.lexpos+1]

def lookahead_n(t, n):
    # TODO what if n is too big?
    return t.lexer.lexdata[t.lexer.lexpos:t.lexer.lexpos+n]

def lookahead_for(t, word):
    return lookahead_n(t, len(word)) == word

def lookahead_for_any(t, words):
    any_found = False
    for word in words:
        if lookahead_for(t, word):
            any_found = True
            break
    return any_found

# Lexer tokens for idio state
def t_idio_AT(t):
    r'@'
    return t

def t_idio_AMP(t):
    r'&'
    return t

def t_idio_COLONS(t):
    r':+'
    return t

def t_idio_DBLQUOTE(t):
    r'"'
    return t

def t_idio_SGLQUOTE(t):
    r'\''
    return t

def t_idio_EXP(t):
    r'export|section'
    return t

def t_idio_END(t):
    r'end'
    return t

def t_idio_WHITESPACE(t):
    r'(\ |\t)+'
    return t

def t_idio_NEWLINE(t):
    r'\r\n|\n|\r'
    exit_idio_state(t)
    return t

def t_idio_IDIOCLOSE(t):
    r'(-->)|(\*/)'
    if not t.lexer.idio_expect_closing_block:
        raise UserFeedback("Unexpected code %s in an idio block" % t.value)
    return t

def t_idio_WORD(t):
    r'[0-9a-zA-Z-_]+'
    return t

def t_idio_OTHER(t):
    r'.'
    t.type = 'CODE'
    exit_idio_state(t)
    return t

def exit_idio_state(t):
    if not t.lexer.idio_expect_closing_block:
        t.lexer.pop_state()
    t.lexer.pop_state()

# Lexer tokens and helpers for idiostart state
def idiostart_incr_comment(t):
    if t.lexer.comment_char == t.value:
        t.lexer.comment_char_count += 1
    else:
        return idiostart_abort(t)

def idiostart_abort(t):
    t.value = t.lexer.lexdata[t.lexer.comment_start_pos:t.lexer.lexpos]
    t.type = "CODE"
    t.lexer.pop_state()
    return t

def t_idiostart_COMMENT(t):
    r'\#|%|;|/|C|!'
    return idiostart_incr_comment(t)

def t_idiostart_SPACE(t):
    r'\ +'
    if t.lexer.comment_char_count != 3:
        return idiostart_abort(t)
    else:   
        t.lexer.push_state('idio')
        t.value = t.lexer.lexdata[t.lexer.comment_start_pos:t.lexer.lexpos]
        t.type = 'IDIO'
        return t

def t_idiostart_ABORT(t):
    r'[^#;/% ]'
    return idiostart_abort(t)

# Lexer tokens and helpers for initial state
def start_idiostart(t):
    if next_char(t) == '\n':
        t.type = 'CODE'
        return t
    else:
        t.lexer.comment_char = t.value
        t.lexer.comment_char_count = 1
        t.lexer.comment_start_pos = t.lexer.lexpos - 1
        t.lexer.idio_expect_closing_block = False
        t.lexer.push_state('idiostart')

def t_IDIOOPEN(t):
    r'(<!--|/\*\*\*)\ +@?'
    if lookahead_for_any(t, ['export', 'section', 'end']):
        t.lexer.push_state('idio')
        t.lexer.idio_expect_closing_block = True
        return t
    else:
        t.type = 'CODE'
        return t

def t_COMMENT(t):
    r'\#|%|;|/|C|!'
    return start_idiostart(t)

def t_NEWLINE(t):
    r'\r\n|\n|\r'
    return t

def t_WHITESPACE(t):
    r'[\ \t]+'
    return t

def t_CODE(t):
    r'[^\#/\n\r]+'
    return t

def p_main(p):
    '''entries : entries entry
               | entry'''
    pass

def p_entry(p):
    '''entry : NEWLINE
             | falsestart
             | codes NEWLINE
             | codes inlineidio NEWLINE
             | idioline NEWLINE'''
    p.lexer.lineno += 1
    if len(p) == 2:
        append_text(p.lexer, p[1])
    elif len(p) == 3:
        if p[1]:
            append_text(p.lexer, p[1] + '\n')
        pass
    elif len(p) == 4:
        code_content = p[1]
        append_text(p.lexer, code_content + "\n")
        # TODO Process inlineidio directives @elide &tag
        # inlineidio_content = p[2]
    else:
        raise Exception("unexpected length " + len(p))

def p_sectionfalsestart(p):
    '''falsestart : IDIO words NEWLINE
                  | IDIO words IDIO NEWLINE
                  | IDIO quote words quote NEWLINE
                  | codes IDIO anythings NEWLINE
                  | WHITESPACE IDIO anythings IDIO NEWLINE
                  | WHITESPACE IDIO anythings NEWLINE'''
    p[0] = "".join(p[1:])

def p_anythings(p):
    '''anythings : anythings anything
                 | anything'''
    p[0] = ''.join(p[1:len(p)])

def p_anything(p):
    '''anything : WORD
                | WHITESPACE
                | CODE'''
    p[0] = p[1]

def p_codes(p):
    '''codes : codes codon
             | codon'''
    if len(p) == 2:
        p[0] = p[1]
    elif len(p) == 3:
        if p[1]:
            p[0] = p[1] + p[2]
        else:
            p[0] = p[2]

def p_codon(p):
    '''codon : CODE 
             | WHITESPACE'''
    p[0] = p[1]

def p_inlineidio(p):
    '''inlineidio : IDIO AMP WORD'''
    p[0] = p[1] + p[2] + p[3]

def p_idioline(p):
    '''idioline : idio
                | idio WHITESPACE
                | WHITESPACE idio
                | WHITESPACE idio WHITESPACE'''
    pass

def p_linecontent(p):
    '''idio : export
            | exportq
            | exportql
            | sectionstart
            | closedcomment
            | closedcommentlevels
            | closedcommentq
            | closedcommentql
            | end '''
    pass

## Methods Defining Section Boundaries
def p_sectionstart(p):
    '''sectionstart : IDIO quote WORD quote
                    | IDIO quote COLONS WORD quote'''
    if len(p) == 5:
        # no colons, so level is 0
        start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[3])
    elif len(p) == 6:
        start_new_section(p.lexer, p.lexpos(1), p.lineno(1), len(p[3]), p[4])
    else:
        raise Exception("unexpected length %s" % len(p))

def p_closed_comment(p):
    '''closedcomment : IDIOOPEN EXP WHITESPACE WORD IDIOCLOSE
                     | IDIOOPEN EXP WHITESPACE WORD WHITESPACE IDIOCLOSE'''
    assert len(p) in [6,7]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[4])

def p_closed_comment_levels(p):
    '''closedcommentlevels : IDIOOPEN EXP WHITESPACE COLONS WORD IDIOCLOSE
                           | IDIOOPEN EXP WHITESPACE COLONS WORD WHITESPACE IDIOCLOSE'''
    assert len(p) in [7,8]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), len(p[4]), p[5])

def p_closed_comment_quoted(p):
    '''closedcommentq : IDIOOPEN EXP WHITESPACE quote words quote IDIOCLOSE
                      | IDIOOPEN EXP WHITESPACE quote words quote WHITESPACE IDIOCLOSE'''
    assert len(p) in [8,9]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[5])

def p_closed_comment_quoted_with_language(p):
    '''closedcommentql : IDIOOPEN EXP WHITESPACE quote words quote WHITESPACE WORD IDIOCLOSE
                       | IDIOOPEN EXP WHITESPACE quote words quote WHITESPACE WORD WHITESPACE IDIOCLOSE'''
    assert len(p) in [10,11]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[5])

## Old Idiopidae @export syntax
def p_export(p):
    '''export : IDIO AT EXP WHITESPACE words
              | IDIO AT EXP WHITESPACE words WHITESPACE'''
    assert len(p) in [6,7]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[5])

def p_export_quoted(p):
    '''exportq : IDIO AT EXP WHITESPACE quote words quote
               | IDIO AT EXP WHITESPACE quote words quote WHITESPACE'''
    assert len(p) in [8,9]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[6])

def p_export_quoted_with_language(p):
    '''exportql : IDIO AT EXP WHITESPACE quote words quote WHITESPACE words
                | IDIO AT EXP WHITESPACE quote words quote WHITESPACE words WHITESPACE'''
    assert len(p) in [10,11]
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), 0, p[6])

def p_end(p):
    '''end : IDIO AT END
           | IDIOOPEN END IDIOCLOSE
           | IDIOOPEN END WHITESPACE IDIOCLOSE
           | IDIOOPEN EXP WHITESPACE quote END quote WHITESPACE IDIOCLOSE'''
    start_new_section(p.lexer, p.lexpos(1), p.lineno(1), p.lexer.level)

def p_quote(p):
    '''quote : DBLQUOTE
             | SGLQUOTE'''
    p[0] = p[1]

def p_words(p):
    '''words : words WHITESPACE WORD
             | WORD'''
    if len(p) == 4:
        p[0] = p[1] + p[2] + p[3]
    elif len(p) == 2:
        p[0] = p[1]
    else:
        raise Exception("unexpected length %s" % len(p))

def p_error(p):
    if not p:
        raise ParseError("Reached EOF when parsing file using idioipdae.")

    lines = p.lexer.lexdata.splitlines()
    this_line = lines[p.lineno-1]

    # Add whole line.
    append_text(p.lexer, this_line+"\n")

    # Forward input to end of line
    while 1:
        tok = yacc.token()
        if not tok or tok.type == 'NEWLINE': break

    yacc.restart()

def tokenize(text, lexer):
    """
    Return array of lexed tokens (for debugging).
    """
    lexer.input(text)
    tokens = []
    while True:
        tok = lexer.token()
        if not tok: break      # No more input
        tokens.append(tok)
    return tokens

def token_info(text, lexer):
    """
    Returns debugging information about lexed tokens as a string.
    """
    def tok_info(tok):
        return "%03d %-15s %s" % (tok.lexpos, tok.type, tok.value.replace("\n",""))
    return "\n".join(tok_info(tok) for tok in tokenize(text, lexer))

# This is outside of the wrapper system so we aren't aware of user-specified
# artifacts directory. Just use .dexy if it's there and don't write files if not.

import os
if os.path.exists(".dexy"):
    outputdir=".dexy"
    lexer = lex.lex(optimize=1, lextab="id_lextab", outputdir=outputdir)
    parser = yacc.yacc(tabmodule="id_parsetab",debug=0, outputdir=outputdir)
else:
    lexer = lex.lex(optimize=0)
    parser = yacc.yacc(write_tables=0, debug=0)



import dexy.exceptions
import dexy.filter
import dexy.node
import os
import pickle
import shutil
import stat
import time

class Doc(dexy.node.Node):
    """
    A single Dexy document.
    """
    aliases = ['doc']
    _settings = {
            'contents' : (
                "Custom contents for a virtual document.",
                None
                ),
            'ws-template' : (
                "custom website template to apply.",
                None
                ),
            'data-type' : (
                "Alias of custom data class to use to store document content.",
                None
                ),
            'shortcut' : (
                "A nickname for document so you don't have to use full key.",
                None
                ),
            'title' : (
                "Custom title.",
                None
                ),
            'output-name' : (
                "Override default canonical name.",
                None
                ),
            'output' : (
                "Whether document should be included in output/ and output-site/",
                None
                )
            }

    def name_args(self):
        name_args = self.setting_values()
        name_args['name'] = self.name
        name_args['rootname'] = os.path.splitext(self.name)[0]
        name_args['basename'] = os.path.basename(self.name)
        name_args['baserootname'] = os.path.splitext(os.path.basename(self.name))[0]
        name_args['dirname'] = os.path.dirname(self.name)
        name_args.update(self.safe_setting('environment', {}))
        return name_args

    def setup_output_name(self):
        """
        Applies string interpolation to %(foo)s settings in output name.
        """
        if not self.setting('output-name'):
            return

        name_args = self.name_args()
        output_name = self.setting('output-name')

        self.log_debug("Name interpolation variables:")
        for key, value in name_args.items():
            self.log_debug("%s: %s" % (key, value))

        if not "/" in output_name:
            output_name = os.path.join(os.path.dirname(self.name), output_name)
        elif output_name.startswith("/"):
            output_name = output_name.lstrip("/")

        try:
            if '%' in output_name:
                updated_output_name = output_name % name_args
            elif '{' in self.setting('output-name'):
                updated_output_name = output_name.format(**name_args)
            else:
                updated_output_name = output_name

        except KeyError as e:
            msg = "Trying to process %s but '%s' is not a valid key. Valid keys are: %s"
            msgargs = (output_name, str(e), ", ".join(sorted(name_args)))
            raise dexy.exceptions.UserFeedback(msg % msgargs)

        self.update_settings({'output-name' : updated_output_name})

    def setup(self):
        self.update_settings(self.args)

        self.name = self.key.split("|")[0]
        self.ext = os.path.splitext(self.name)[1]
        self.filter_aliases = self.key.split("|")[1:]
        self.filters = []

        self.setup_output_name()
        self.setup_initial_data()

        for alias in self.filter_aliases:
            f = dexy.filter.Filter.create_instance(alias, self)
            self.filters.append(f)

        prev_filter = None
        for i, f in enumerate(self.filters):
            filter_aliases = self.filter_aliases[0:i+1]
            filter_key = "%s|%s" % (self.name, "|".join(filter_aliases))
            storage_key = "%s-%03d-%s" % (self.hashid, i+1, "-".join(filter_aliases))

            if i < len(self.filters) - 1:
                next_filter = self.filters[i+1]
            else:
                next_filter = None

            filter_settings_from_args = self.args.get(f.alias, {})
            f.setup(filter_key, storage_key, prev_filter,
                    next_filter, filter_settings_from_args)
            prev_filter = f

    def setup_datas(self):
        """
        Convenience function to ensure all datas are set up. Should not need to be called normally.
        """
        for d in self.datas():
            if d.state == 'new':
                d.setup()

    def setup_initial_data(self):
        storage_key = "%s-000" % self.hashid

        if self.setting('output') is not None:
            canonical_output = self.setting('output')
        else:
            if len(self.filter_aliases) == 0:
                canonical_output = True
            else:
                canonical_output = None

        settings = {
                'canonical-name' : self.name,
                'canonical-output' : canonical_output,
                'shortcut' : self.setting('shortcut'),
                'output-name' : self.setting('output-name'),
                'title' : self.setting('title')
                }

        self.initial_data = dexy.data.Data.create_instance(
                self.data_class_alias(),
                self.name, # key
                self.ext, #ext
                storage_key,
                settings,
                self.wrapper
                )

    def consolidate_cache_files(self):
        for node in self.input_nodes():
            node.consolidate_cache_files()

        if self.state == 'cached':
            self.setup_datas()

            # move cache files to new cache
            for d in self.datas():
                if os.path.exists(d.storage.last_data_file()):
                    last_loc = d.storage.last_data_file()
                    this_loc = d.storage.this_data_file()

                    self.log_debug("moving %s from %s to %s..." % (d.key, last_loc, this_loc))
                    shutil.move(last_loc, this_loc)

            if os.path.exists(self.runtime_info_filename(False)):
                shutil.move(self.runtime_info_filename(False), self.runtime_info_filename(True))

            self.apply_runtime_info()

            for d in self.datas():
                if hasattr(d.storage, 'connect'):
                    d.storage.connect()
            self.transition('consolidated')

    def apply_runtime_info(self):
            runtime_info = self.load_runtime_info()
            if runtime_info:
                self.add_runtime_args(runtime_info['runtime-args'])
                self.load_additional_docs(runtime_info['additional-docs'])

    def datas(self):
        """
        Returns all associated `data` objects.
        """
        return [self.initial_data] + [f.output_data for f in self.filters]

    def update_setting(self, key, value):
        self.update_all_settings({key : value})

    def update_all_settings(self, new_settings):
        self.update_settings(new_settings)

        for data in self.datas():
            data.update_settings(new_settings)

        for f in self.filters:
            f.update_settings(new_settings)

    def check_cache_elements_present(self):
        """
        Returns a boolean to indicate whether all files are present in cache.
        """
        # Take this opportunity to ensure Data objects are in `setup` state.
        for d in self.datas():
            if d.state == 'new':
                d.setup()

        return all(
                os.path.exists(d.storage.last_data_file()) or
                os.path.exists(d.storage.this_data_file())
                for d in self.datas())

    def check_doc_changed(self):
        if self.name in self.wrapper.filemap:
            live_stat = self.wrapper.filemap[self.name]['stat']

            self.initial_data.setup()

            in_this_cache = os.path.exists(self.initial_data.storage.this_data_file())
            in_last_cache = os.path.exists(self.initial_data.storage.last_data_file())

            if in_this_cache or in_last_cache:
                # we have a file in the cache from a previous run, compare its
                # mtime to filemap to determine whether it has changed
                if in_this_cache:
                    cache_stat = os.stat(self.initial_data.storage.this_data_file())
                else:
                    cache_stat = os.stat(self.initial_data.storage.last_data_file())

                cache_mtime = cache_stat[stat.ST_MTIME]
                live_mtime = live_stat[stat.ST_MTIME]
                msg = "    cache mtime %s live mtime %s now %s changed (live gt cache) %s"
                msgargs = (cache_mtime, live_mtime, time.time(), live_mtime > cache_mtime)
                self.log_debug(msg % msgargs)
                return live_mtime > cache_mtime
            else:
                # there is no file in the cache, therefore it has 'changed'
                return True
        else:
            # TODO check hash of contents of virtual files
            return False

    def data_class_alias(self):
        data_class_alias = self.setting('data-type')

        if data_class_alias:
            return data_class_alias
        else:
            contents = self.get_contents()
            if isinstance(contents, dict):
                return 'keyvalue'
            elif isinstance(contents, list):
                return "sectioned"
            else:
                return 'generic'

    def get_contents(self):
        contents = self.setting('contents')
        return contents

    # Runtime Info
    def runtime_info_filename(self, this=True):
        name = "%s.runtimeargs.pickle" % self.hashid
        return os.path.join(self.initial_data.storage.storage_dir(this), name)

    def save_runtime_info(self):
        """
        Save runtime changes to metadata so they can be reapplied when node has
        been cached.
        """

        info = {
            'runtime-args' : self.runtime_args,
            'additional-docs' : self.additional_doc_info()
            }

        with open(self.runtime_info_filename(), 'wb') as f:
            pickle.dump(info, f)

    def load_runtime_info(self):
        info = None

        # Load from 'this' first
        try:
            with open(self.runtime_info_filename(), 'rb') as f:
                info = pickle.load(f)
        except IOError:
            pass

        # Load from 'last' if there's nothing in 'this'
        if not info:
            try:
                with open(self.runtime_info_filename(False), 'rb') as f:
                    info = pickle.load(f)
            except IOError:
                pass

        return info

    def run(self):
        if self.wrapper.directory != '.':
            if not self.wrapper.directory in self.name:
                print(("skipping", self.name, "not in", self.wrapper.directory))
                return

        self.start_time = time.time()

        if self.name in self.wrapper.filemap:
            # This is a real file on the file system.
            if self.doc_changed or not self.initial_data.is_cached():
                self.initial_data.copy_from_file(self.name)
        else:
            is_dummy = self.initial_data.is_cached() and self.get_contents() == 'dummy contents'
            if is_dummy:
                self.initial_data.load_data()
            else:
                self.initial_data.set_data(self.get_contents())

        for f in self.filters:
            f.start_time = time.time()
            if f.output_data.state == 'new':
                f.output_data.setup()
            if hasattr(f.output_data.storage, 'connect'):
                f.output_data.storage.connect()
            f.process()
            f.finish_time = time.time()
            f.elapsed = f.finish_time - f.start_time

        self.finish_time = time.time()
        self.elapsed_time = self.finish_time - self.start_time
        self.wrapper.batch.add_doc(self)
        self.save_runtime_info()

        # Run additional docs
        for doc in self.additional_docs:
            doc.check_is_cached()
            for task in doc:
                task()

    def output_data(self):
        """
        Returns a reference to the final data object for this document.
        """
        if self.filters:
            return self.filters[-1].output_data
        else:
            return self.initial_data

    def batch_info(self):
        return {
                'input-data' : self.initial_data.args_to_data_init(),
                'output-data' : self.output_data().args_to_data_init(),
                'filters-data' : [f.output_data.args_to_data_init() for f in self.filters],
                # below are convenience attributes, not strictly necessary for dexy to run
                'title' : self.output_data().title(),
                'start_time' : self.start_time,
                'finish_time' : self.finish_time,
                'elapsed' : self.elapsed_time,
                'state' : self.state
                }

    def finalize_additional_doc(self):
        """
        Fixes 'contents' setting of additional docs where initial content is
        set after creation.
        """
        self.update_settings({"contents" : self.output_data().data()})



from dexy.exceptions import UserFeedback
from dexy.parser import AbstractSyntaxTree
from dexy.parsers.doc import Yaml
from tests.utils import capture_stderr
from tests.utils import wrap
from dexy.wrapper import Wrapper
import dexy.batch
import os

def test_hard_tabs_in_config():
    with wrap():
        with capture_stderr() as stderr:
            os.makedirs("abc/def")
            with open("abc/def/dexy.yaml", "w") as f:
                f.write("""foo:\t- .txt""")

            wrapper = Wrapper()

            try:
                wrapper.run_from_new()
            except UserFeedback as e:
                assert "hard tabs" in str(e)

            assert "abc/def/dexy.yaml" in stderr.getvalue()

def test_subdir_config_with_bundle():
    with wrap():
        with open("dexy.yaml", "w") as f:
            f.write("""
            foo:
                - .txt
            """)

        os.makedirs("abc/def")
        with open("abc/def/dexy.yaml", "w") as f:
            f.write("""
            bar:
                - .py
            """)

        with open("abc/def/hello.py", "w") as f:
            f.write("print 'hello'")

        wrapper = Wrapper()
        wrapper.run_from_new()
        assert "doc:abc/def/hello.py" in wrapper.nodes

        wrapper = Wrapper(recurse=False)
        wrapper.run_from_new()
        assert not "doc:abc/def/hello.py" in wrapper.nodes

        wrapper = Wrapper(recurse=False, configs="abc/def/dexy.yaml")
        wrapper.run_from_new()
        assert "doc:abc/def/hello.py" in wrapper.nodes

def test_except_patterndoc():
    with wrap() as wrapper:
        with open("exceptme.abc", "w") as f:
            f.write("hello")

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', """.abc:\n  - except : 'exceptme.abc' """)
        ast.walk()

        assert len(wrapper.nodes) == 1

def test_except_patterndoc_pattern():
    with wrap() as wrapper:
        with open("exceptme.abc", "w") as f:
            f.write("hello")

        wrapper = Wrapper()
        wrapper.to_valid()

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', """.abc:\n  - except : 'exceptme.*' """)
        ast.walk()
        wrapper.transition('walked')
        wrapper.to_checked()

        wrapper.run()

        assert len(wrapper.nodes) == 1

def test_children_siblings_order():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', """
        p1:
            - c1
            - c2:
                - g1
                - g2
                - g3
            - c3
        """)
        ast.walk()
        wrapper.transition('walked')

        wrapper.to_checked()

        wrapper.run()

        p1 = wrapper.nodes['bundle:p1']
        assert [i.key_with_class() for i in p1.walk_inputs()] == [
                'bundle:c1',
                'bundle:c2',
                'bundle:g1',
                'bundle:g2',
                'bundle:g3',
                'bundle:c3'
                ]

        c1 = wrapper.nodes['bundle:c1']
        assert len(c1.inputs) == 0

        c2 = wrapper.nodes['bundle:c2']
        assert [i.key_with_class() for i in c2.walk_inputs()] == [
                'bundle:g1',
                'bundle:g2',
                'bundle:g3'
                ]

        c3 = wrapper.nodes['bundle:c3']
        assert len(c3.inputs) == 0

        g3 = wrapper.nodes['bundle:g3']
        assert len(g3.inputs) == 0

def test_single_file_doc():
    with wrap() as wrapper:
        with open("hello.txt", "w") as f:
            f.write("hello")

        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', "hello.txt")
        ast.walk()
        wrapper.transition('walked')

        wrapper.to_checked()

        wrapper.run()
        assert "doc:hello.txt" in wrapper.nodes

def test_single_bundle_doc():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', "hello")
        ast.walk()
        wrapper.transition('walked')
        wrapper.to_checked()
        assert "bundle:hello" in wrapper.nodes

def test_single_bundle_doc_with_args():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', """
        more:
            - hello
            - one-more-task
            - foo: bar

        hello:
            - foo: bar
            - filter_fruit: orange
            - args:
                - ping: pong
            - another-task:
                - foo: baz
                - yet-another-task:
                    - foo: bar
            - one-more-task
        """)
        ast.walk()

        assert wrapper.roots[0].key_with_class() == "bundle:more"
        assert len(wrapper.nodes) == 5

def test_single_bundle_doc_with_args_2():
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', """

      -  hello:
            - foo: bar
            - filter_fruit: orange
            - args:
                - ping: pong
            - another-task:
                - foo: baz
                - yet-another-task:
                    - foo: bar
            - one-more-task

      -  more:
            - hello
            - one-more-task
            - foo: bar

        """)

        ast.walk()

        assert wrapper.roots[0].key_with_class() == "bundle:more"
        assert len(wrapper.nodes) == 5



from tests.utils import runfilter

def test_nested_html():
    with runfilter('soups', nested_html) as doc:
        data = doc.output_data()
        assert str(data) == expected
        assert list(data.keys()) == ['First', 'Second', 'Actual Document Contents']

def test_markdown_output():
    with runfilter('markdown|soups', md) as doc:
        data = doc.output_data()

        assert list(data.keys()) == ['foo', 'bar', 'barbaz', 'Actual Document Contents']

        assert data['foo']['level'] == 1
        assert data['bar']['level'] == 1
        assert data['barbaz']['level'] == 2

        assert data['foo']['id'] == 'foo'
        assert data['bar']['id'] == 'bar'
        assert data['barbaz']['id'] == 'barbaz'

def test_soup_sections_filter():
    with runfilter('soups', html, ext='.html') as doc:
        data = doc.output_data()

        assert list(data.keys()) == ['The First Named Section',
                'Nested In First Section', 'The 2nd Section',
                'Actual Document Contents']

        first_section = data["The First Named Section"]
        assert first_section['contents'] == None
        assert first_section['level'] == 1

        nested_section = data["Nested In First Section"]
        assert nested_section['level'] == 2
        assert first_section['contents'] == None

        final_section = data["The 2nd Section"]
        assert final_section['level'] == 1
        assert final_section['contents'] == None

        contents_section = data['Actual Document Contents']
        assert contents_section['level'] == 1

def test_no_blank_anonymous_first_section():
    with runfilter('soups', "<h1>first</h1><p>foo</p><h1>second</h1>", ext=".html") as doc:
        assert list(doc.output_data().keys()) == ['first', 'second', 'Actual Document Contents']

nested_html = """<div>
<h1>First</h1>
<div>
<h2>Second</h2>
</div>
</div>"""

expected = """<div>
<h1 id="first">First</h1>
<div>
<h2 id="second">Second</h2>
</div>
</div>"""

md = """# foo

This is the foo section.

# bar

This is the bar section.

## barbaz

This is the barbaz section.

"""

html = """
<p>Text before the first section</p>
<H1>The First Named Section</H1>
<p>Some content in the first section.</p>
<h2>Nested In First Section</h2>
<p>Content in the nested section.</p>
<ul>
<li>list item the first</li>
<li>list item the second</li>
</ul>
<h1>The 2nd Section</h1>
<p>foo.</p>
"""



from dexy.doc import Doc
from dexy.filters.process import SubprocessFilter
from tests.utils import wrap
import dexy.exceptions
import os

def test_add_new_files():
    with wrap() as wrapper:
        node = Doc("example.sh|sh",
                wrapper,
                [],
                contents = "echo 'hello' > newfile.txt",
                sh = {
                    "add-new-files" : True,
                    "keep-originals" : True,
                    "additional-doc-filters" : { '.txt' : 'markdown' }
                    }
                )
        wrapper.run_docs(node)

        assert str(wrapper.nodes['doc:newfile.txt'].output_data()) == "hello" + os.linesep
        assert str(wrapper.nodes['doc:newfile.txt|markdown'].output_data()) == "<p>hello</p>"

def test_not_present_executable():
    # TODO modify test so we try to run this
    dexy.filter.Filter.create_instance('notreal')

class NotPresentExecutable(SubprocessFilter):
    """
    notreal
    """
    EXECUTABLE = 'notreal'
    aliases = ['notreal']

def test_command_line_args():
    with wrap() as wrapper:
        node = Doc("example.py|py",
                wrapper,
                [],
                py={"args" : "-B"},
                contents="print('hello')"
                )
        wrapper.run_docs(node)

        assert str(node.output_data()) == "hello" + os.linesep

        command_used = node.filters[-1].command_string()
        assert command_used == "python -B  \"example.py\" "

def test_scriptargs():
    with wrap() as wrapper:
        node = Doc("example.py|py",
                wrapper,
                [],
                py={"scriptargs" : "--foo"},
                contents="""import sys\nprint("args are: '%s'" % sys.argv[1])"""
                )
        wrapper.run_docs(node)

        assert "args are: '--foo'" in str(node.output_data())

        command_used = node.filters[-1].command_string()
        assert command_used == "python   \"example.py\" --foo"

def test_custom_env_in_args():
    with wrap() as wrapper:
        node = Doc("example.py|py",
                wrapper,
                [],
                py={"env" : {"FOO" : "bar" }},
                contents="import os\nprint(os.environ['FOO'])"
                )
        wrapper.run_docs(node)

        assert str(node.output_data()) == "bar" + os.linesep

def test_nonzero_exit():
    with wrap() as wrapper:
        wrapper.debug = False
        node = Doc("example.py|py",
                wrapper,
                [],
                contents="import sys\nsys.exit(1)"
                )
        wrapper.run_docs(node)
        assert wrapper.state == 'error'

def test_ignore_nonzero_exit():
    with wrap() as wrapper:
        wrapper.ignore_nonzero_exit = True
        node = Doc("example.py|py",
                wrapper,
                [],
                contents="import sys\nsys.exit(1)"
                )
        wrapper.run_docs(node)
        assert True # no NonzeroExit was raised...
